#!/usr/bin/env python
"""Tests for `geomeffibem` Surface class."""

import numpy as np
import openstudio
import pytest

from geomeffibem.vertex import Vertex, distanceFromPointToLine, isPointOnLineBetweenPoints


def test_vertex_from_numpy():
    """Vertex from numpy."""
    arr = np.array([0.1, 0.2, 0.3])
    v = Vertex.from_numpy(arr)
    assert v.x == arr[0]
    assert v.y == arr[1]
    assert v.z == arr[2]
    with pytest.raises(ValueError):
        Vertex.from_numpy(np.array([1.0, 2.0, 3.0, 4.0]))


def test_vertex_from_list():
    """Vertex from a list."""
    mylist = [0.1, 0.2, 0.3]
    v = Vertex.from_numpy(mylist)
    assert v.x == mylist[0]
    assert v.y == mylist[1]
    assert v.z == mylist[2]


def test_vertex_to_numpy():
    """To numpy."""
    v = Vertex(0.1, 0.2, 0.3)
    arr = v.to_numpy()
    assert isinstance(arr, np.ndarray)
    assert arr.shape == (3,)
    assert v.x == arr[0]
    assert v.y == arr[1]
    assert v.z == arr[2]


def test_vertex_from_Point3d():
    """From openstudio Point3d."""
    pt = openstudio.Point3d(0.1, 0.2, 0.3)
    v = Vertex.from_Point3d(pt)
    assert v.x == pt.x()
    assert v.y == pt.y()
    assert v.z == pt.z()


def test_vertex_to_Point3d():
    """To openstudio Point3d."""
    v = Vertex(0.1, 0.2, 0.3)
    pt = v.to_Point3d()
    assert v.x == pt.x()
    assert v.y == pt.y()
    assert v.z == pt.z()


def test_vertex_equality():
    """Tests equality with a tolerance."""
    # Baked in tolerance is 0.0127 to match E+
    assert Vertex(0.0, 0.0, 0.0) == Vertex(0.0126, 0.0, 0.0)
    assert not (Vertex(0.0, 0.0, 0.0) != Vertex(0.0126, 0.0, 0.0))

    assert Vertex(0.0, 0.0, 0.0) != Vertex(0.0128, 0.0, 0.0)
    assert not (Vertex(0.0, 0.0, 0.0) == Vertex(0.0128, 0.0, 0.0))


def test_coords_on_plane():
    """Tests the Vertex.get_coords_on_plane."""
    v = Vertex(1.0, 2.0, 3.0)
    assert v.get_coords_on_plane('xy') == (1.0, 2.0)
    assert v.get_coords_on_plane('yz') == (2.0, 3.0)
    assert v.get_coords_on_plane('xz') == (1.0, 3.0)
    with pytest.raises(ValueError):
        v.get_coords_on_plane('dkjngfdjkn')


def test_distanceFromPointToLine():
    """Tests free function distanceFromPointToLine."""
    start = Vertex(0.0, 0.0, 0.0)
    end = Vertex(10.0, 0.0, 0.0)
    testVertex = Vertex(5.0, 0.0, 0.0)
    assert distanceFromPointToLine(start=start, end=end, test=testVertex) == 0.0
    testVertex.y = 5.0
    assert distanceFromPointToLine(start=start, end=end, test=testVertex) == 5.0


def test_isPointOnLineBetweenPoints():
    """Test free function isPointOnLineBetweenPoints."""
    start = Vertex(0.0, 0.0, 0.0)
    end = Vertex(10.0, 0.0, 0.0)
    testVertex = Vertex(5.0, 0.0, 0.0)
    assert isPointOnLineBetweenPoints(start=start, end=end, test=testVertex)
    testVertex.x = 11.0
    assert not isPointOnLineBetweenPoints(start=start, end=end, test=testVertex)
    testVertex.x = 5.0
    testVertex.y = 0.0126
    assert isPointOnLineBetweenPoints(start=start, end=end, test=testVertex)

    testVertex.y = 1.0
    assert not isPointOnLineBetweenPoints(start=start, end=end, test=testVertex)



#!/usr/bin/env python
"""Tests for `geomeffibem` Surface class."""

import matplotlib.pyplot as plt
import numpy as np
import openstudio
import pytest

from geomeffibem.surface import Surface, Surface3dEge, get_surface_from_surface_like, plot_vertices
from geomeffibem.vertex import Vertex


def test_surface_rectangle():
    """Tests Surface.Rectangle factory."""
    z = 0.0
    surface = Surface.Rectangle(min_x=0.0, max_x=10.0, min_y=0.0, max_y=10.0, min_z=z, max_z=z)
    assert np.array_equal(
        surface.to_numpy(),
        np.array(
            [
                [0.0, 10.0, z],
                [0.0, 0.0, z],
                [10.0, 0.0, z],
                [10.0, 10.0, z],
            ]
        ),
    )
    assert surface.get_plot_axis() == 'xy'

    surface = Surface.Rectangle(min_x=0.0, max_x=10.0, min_y=5.0, max_y=5.0, min_z=10.0, max_z=20.0)
    assert np.array_equal(
        surface.to_numpy(),
        np.array(
            [
                [0.0, 5.0, 20.0],
                [0.0, 5.0, 10.0],
                [10.0, 5.0, 10.0],
                [10.0, 5.0, 20.0],
            ]
        ),
    )

    assert surface.get_plot_axis() == 'xz'

    surface_rot = surface.rotate(-15)
    with pytest.raises(NotImplementedError):
        surface_rot.get_plot_axis()


def test_surface_roundtrip_openstudio():
    """Tests to/from openstudio.model.Surface."""
    surface = Surface.Rectangle(min_x=0.0, max_x=10.0, min_y=0.0, max_y=10.0, min_z=0.0, max_z=0.0)
    assert surface.os_area() == 100.0
    m = openstudio.model.Model()
    sf = openstudio.model.Surface(surface.to_Point3dVector(), m)
    assert isinstance(sf, openstudio.openstudiomodelgeometry.Surface)
    mysf = Surface.from_Surface(sf)
    assert np.array_equal(surface.to_numpy(), mysf.to_numpy())

    mysf2 = Surface.from_Point3dVector(mysf.to_Point3dVector())
    assert np.array_equal(surface.to_numpy(), mysf2.to_numpy())
    with pytest.raises(ValueError):
        Surface.from_Surface(mysf.to_numpy())


def test_surface_roundtrip_numpy():
    """Tests to/from numpy."""
    surface = Surface.Rectangle(min_x=0.0, max_x=10.0, min_y=0.0, max_y=10.0, min_z=0.0, max_z=0.0)
    surface2 = Surface.from_numpy_array(surface.to_numpy())
    assert np.array_equal(surface.to_numpy(), surface2.to_numpy())

    # with a list
    Surface.from_numpy_array(surface.to_numpy().tolist())

    with pytest.raises(ValueError):
        Surface.from_numpy_array(np.array([[0.0, 0.0, 0.0], [0.0, 10.0, 0.0]]))

    with pytest.raises(ValueError):
        Surface.from_numpy_array(np.array([[0.0, 0.0, 0.0, 0.0], [0.0, 10.0, 0.0, 0.0], [10.0, 10.0, 0.0, 0.0]]))

    with pytest.raises(ValueError):
        Surface(vertices=surface)

    with pytest.raises(ValueError):
        Surface(vertices=surface.to_Point3dVector())


def test_Surface3dEdge():
    """Test Surface3dEge."""
    surface = Surface.Rectangle(min_x=0.0, max_x=10.0, min_y=0.0, max_y=10.0, min_z=0.0, max_z=0.0)
    surface.name = 'Floor'
    edge = Surface3dEge(start=Vertex(0.0, 0.0, 0.0), end=Vertex(0.0, 10.0, 0.01), firstSurface=surface)
    assert edge == surface.to_Surface3dEdges()[0]

    assert edge.containsPoints(Vertex(0.0, 1.0, 0.0))
    assert edge.containsPoints(Vertex(0.0, 1.0, 0.01))
    assert not edge.containsPoints(Vertex(0.0, 0.0, 0.0))
    assert not edge.containsPoints(Vertex(0.0, 1.0, 1.0))
    assert edge.count() == 1

    edge = Surface3dEge(start=Vertex(0.0, 0.0, 0.0), end=Vertex(0.0, 10.0, 1.0), firstSurface=surface)
    assert edge != surface.to_Surface3dEdges()[0]

    edge.plot_on_first_surface()

    fig, ax = plt.subplots()
    edge.plot_on_first_surface(ax=ax)


def test_Surface_get_plot_axis():
    """Test Surface.get_plot_axis."""
    floor_surface = Surface.Rectangle(min_x=0.0, max_x=10.0, min_y=0.0, max_y=10.0, min_z=0.0, max_z=0.0)
    assert floor_surface.get_plot_axis() == 'xy'

    yz_wall = Surface.from_numpy_array(np.array([[0.0, 10.0, 0.3], [0.0, 10.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.3]]))
    assert yz_wall.get_plot_axis() == 'yz'

    xz_wall = Surface.from_numpy_array(np.array([[0.0, 0.0, 0.3], [0.0, 0.0, 0.0], [30.0, 0.0, 0.0], [30.0, 0.0, 0.3]]))
    assert xz_wall.get_plot_axis() == 'xz'

    # This should be able to figure out the plane by itself
    floor_surface.plot()
    yz_wall.plot()
    xz_wall.plot()


def test_Surface_centroid():
    """Test the centroid methods."""
    surface = Surface.Rectangle(min_x=0.0, max_x=10.0, min_y=0.0, max_y=10.0, min_z=0.0, max_z=0.0)
    assert surface.os_centroid() == Vertex(+5.0000, +5.0000, +0.0000)
    assert surface.rough_centroid() == Vertex(+5.0000, +5.0000, +0.0000)

    # This isn't a surface, just four equal points...
    surface = Surface.Rectangle(min_x=0.0, max_x=0.0, min_y=0.0, max_y=0.0, min_z=0.0, max_z=0.0)
    with pytest.raises(ValueError):
        surface.os_centroid()


def test_Surface_split():
    """Test splitting a surface."""
    surface = Surface.Floor(min_x=0.0, max_x=10.0, min_y=0.0, max_y=10.0, z=0.0)
    surface.name = "Surface"
    surface.os_centroid() == Vertex(+5.0000, +5.0000, +0.0000)

    s1, s2 = surface.split_into_n_segments(n_segments=2, axis=None, plot=False)
    assert s1.name == "Surface-1"
    assert np.array_equal(
        s1.to_numpy(), np.array([[5.0, 10.0, 0.0], [5.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, 10.0, 0.0]])
    )

    assert s2.name == "Surface-2"
    assert np.array_equal(
        s2.to_numpy(), np.array([[10.0, 10.0, 0.0], [10.0, 0.0, 0.0], [5.0, 0.0, 0.0], [5.0, 10.0, 0.0]])
    )

    with pytest.raises(ValueError):
        surface.split_into_n_segments(n_segments=2, axis='z', plot=False)

    surface.name = None
    s1, s2 = surface.split_into_n_segments(n_segments=2, axis='y', plot=False)
    assert s1.name is None
    assert np.array_equal(
        s1.to_numpy(), np.array([[10.0, 5.0, 0.0], [10.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, 5.0, 0.0]])
    )

    assert s2.name is None
    assert np.array_equal(
        s2.to_numpy(), np.array([[10.0, 10.0, 0.0], [10.0, 5.0, 0.0], [0.0, 5.0, 0.0], [0.0, 10.0, 0.0]])
    )

    with pytest.raises(ValueError):
        surface.split_into_n_segments(n_segments=1, axis='x', plot=False)


def test_split_plot():
    """Test split and plot."""
    surface = Surface.Rectangle(min_x=0.0, max_x=10.0, min_y=0.0, max_y=10.0, min_z=0.0, max_z=0.0)
    surface.name = "Surface"
    surface.split_into_n_segments(n_segments=2, axis='y', plot=True)


def test_Surface_rotate():
    """Test rotate."""
    surface = Surface.Floor(min_x=0.0, max_x=10.0, min_y=0.0, max_y=10.0, z=0.0)

    surface_rot = surface.rotate(degrees=-25.0)
    assert np.isclose(surface.os_area(), surface_rot.os_area())

    surface_rot = surface.rotate(degrees=-90.0)

    assert np.allclose(
        surface_rot.to_numpy(),
        np.array(
            [
                [-10.0, +10.0, +0.0],
                [+0.0, +10.0, +0.0],
                [+0.0, +0.0, +0.0],
                [-10.0, +0.0, +0.0],
            ]
        ),
    )


def test_surface_plot():
    """Test plotting."""
    surface = Surface.Rectangle(min_x=0.0, max_x=10.0, min_y=0.0, max_y=10.0, min_z=0.0, max_z=0.0)
    surface.name = "Floor"
    surface.plot(with_rough_centroid=True, with_os_centroid=True, center_axes=True)

    surface.plot(with_rough_centroid=True, with_os_centroid=False, center_axes=False)
    # no name
    surface.name = None
    surface.plot(with_rough_centroid=True, with_os_centroid=True, center_axes=False)
    surface.plot(with_rough_centroid=True, with_os_centroid=False, center_axes=False)
    surface.plot(name="Custom", with_os_centroid=False, center_axes=False)
    surface.plot(name=False, with_os_centroid=False, center_axes=False)

    with pytest.raises(ValueError):
        plot_vertices(surface, plane='wrong')


def test_get_surface_from_surface_like():
    """Test get_surface_from_surface_like."""
    mylist = [
        [+0.0, +0.0, +0.0],
        [-10.0, +0.0, +0.0],
        [-10.0, +10.0, +0.0],
        [+0.0, +10.0, +0.0],
    ]
    assert isinstance(get_surface_from_surface_like(mylist), Surface)
    assert isinstance(get_surface_from_surface_like(np.array(mylist)), Surface)
    surface = Surface.Rectangle(min_x=0.0, max_x=10.0, min_y=0.0, max_y=10.0, min_z=0.0, max_z=0.0)
    surface.name = "Floor"
    assert isinstance(get_surface_from_surface_like(surface), Surface)
    assert isinstance(get_surface_from_surface_like(surface.to_Point3dVector()), Surface)
    assert isinstance(get_surface_from_surface_like(surface.to_numpy()), Surface)

    m = openstudio.model.Model()
    os_sf = surface.to_OSSurface(m)
    assert isinstance(get_surface_from_surface_like(os_sf), Surface)

    assert isinstance(get_surface_from_surface_like(surface.vertices), Surface)



"""Surface and Surface3dEdge objects.

A Surface is a collection of Vertex.
A Surface3dEdge is a side of a Surface.
"""

from __future__ import annotations

import copy
from typing import List, Union

import matplotlib.pyplot as plt
import numpy as np
import openstudio

from geomeffibem.plane import Plane
from geomeffibem.vertex import (
    Vertex,
    distance,
    getAngle,
    getNewellVector,
    getOutwardNormal,
    isAlmostEqual3dPt,
    isPointOnLineBetweenPoints,
)


class Surface3dEge:
    """An Edge has a start and an end Vertex, and a list of surfaces it was found on."""

    def __init__(self, start: Vertex, end: Vertex, firstSurface: Surface):
        """Constructor."""
        self.start = start
        self.end = end
        self.allSurfaces = [firstSurface]

    def containsPoints(self, testVertex: Vertex) -> bool:
        """Checks whether a Point is on the edge.

        It is not almost equal to the start and end points, and,
        isPointOnLineBetweenPoints(start, end, testVertex) is true.
        """
        return (
            not isAlmostEqual3dPt(self.start, testVertex)
            and not isAlmostEqual3dPt(self.end, testVertex)
            and isPointOnLineBetweenPoints(self.start, self.end, testVertex)
        )

    def length(self) -> float:
        """Compute distance from start to end."""
        return distance(self.start, self.end)

    def count(self) -> int:
        """Number of Surfaces it was found on."""
        return len(self.allSurfaces)

    def __eq__(self, other):
        """Operator equal."""
        if not isinstance(other, Surface3dEge):
            raise NotImplementedError("Not implemented for any other types than Surface3dEge itself")

        return (isAlmostEqual3dPt(self.start, other.start) and isAlmostEqual3dPt(self.end, other.end)) or (
            isAlmostEqual3dPt(self.start, other.end) and isAlmostEqual3dPt(self.end, other.start)
        )

    def __ne__(self, other):
        """Operator not equal."""
        return not self == other

    def __repr__(self):
        """Repr."""
        return f"start={self.start}, end={self.end}, count={self.count()}, firstSurface={self.allSurfaces[0].name}"

    def plot_on_first_surface(self, ax=None):
        """Plots this segment in red atop the outline of the Surface it came from."""
        surface = self.allSurfaces[0]

        if ax is None:
            fig, ax = plt.subplots(figsize=(16, 9))
        surface.plot(ax=ax)
        plot_vertices([self.start, self.end], plane=surface.get_plot_axis(), c='r', ax=ax, annotate=False)


class Surface:
    """A 3D Surface."""

    @staticmethod
    def from_numpy_array(arr) -> Surface:
        """Factory method to construct from a numpy array of 3-coordinates arrays."""
        if isinstance(arr, list):
            arr = np.array(arr)
        if arr.shape[0] < 3:
            raise ValueError("Need at least 3 vertices to construct a Surface")
        if arr.shape[1] != 3:
            raise ValueError(f"Expected a numpy array with a dimension (N, 3), got {arr.shape}")
        return Surface([Vertex.from_numpy(x) for x in arr])

    @staticmethod
    def from_Point3dVector(points: Union[openstudio.Point3dVector, List[openstudio.Point3d]]) -> Surface:
        """Factory method to construct from an openstudio Point3dVector or a list of Point3d."""
        return Surface(vertices=[Vertex.from_Point3d(x) for x in points])

    @staticmethod
    def from_Surface(openstudio_surface: openstudio.model.Surface) -> Surface:
        """Factory method to construct from an openstudio.model.Surface."""
        if not isinstance(openstudio_surface, openstudio.model.Surface):
            raise ValueError("Expected an openstudio.model.Surface")
        return Surface(
            vertices=[Vertex.from_Point3d(x) for x in openstudio_surface.vertices()],
            name=openstudio_surface.nameString(),
        )

    @staticmethod
    def Floor(min_x=0.0, max_x=10.0, min_y=0.0, max_y=10.0, z=0.0) -> Surface:
        """Create a rectangular floor Surface (outward normal pointing down)."""
        # Counterclockwise, ULC convention, except here we want to create a floor so
        # outward normal must be pointing DOWN, so clockwise order
        vertices_arr = np.array(
            [
                [max_x, max_y, z],
                [max_x, min_y, z],
                [min_x, min_y, z],
                [min_x, max_y, z],
            ]
        )
        vertices = [Vertex.from_numpy(x) for x in vertices_arr]

        return Surface(vertices=vertices)

    @staticmethod
    def Rectangle(min_x=0.0, max_x=10.0, min_y=0.0, max_y=10.0, min_z=0.0, max_z=0.0) -> Surface:
        """Factory method to easily create a rectangular Surface, with ULC convention."""
        if abs(max_z - min_z) < 0.01:
            z = min_z
            if abs(z) < 0.01:
                print(
                    "Looks like you're trying to create a Floor surface... "
                    "use the Surface.Floor factory method if that's the case so outwardNormal points down."
                )

            vertices_arr = np.array(
                [
                    [min_x, max_y, z],  # Upper Left Corner
                    [min_x, min_y, z],  # Lower Left Corner
                    [max_x, min_y, z],  # Lower Right Corner
                    [max_x, max_y, z],  # Upper Right Corner
                ]
            )
        elif abs(max_x - min_x) < 0.01:
            x = min_x
            vertices_arr = np.array(
                [
                    [x, min_y, max_z],  # Upper Left Corner
                    [x, min_y, min_z],  # Lower Left Corner
                    [x, max_y, min_z],  # Lower Right Corner
                    [x, max_y, max_z],  # Upper Right Corner
                ]
            )

        elif abs(max_y - max_y) < 0.01:
            y = min_y
            vertices_arr = np.array(
                [
                    [min_x, y, max_z],  # Upper Left Corner
                    [min_x, y, min_z],  # Lower Left Corner
                    [max_x, y, min_z],  # Lower Right Corne
                    [max_x, y, max_z],  # Upper Right Corner
                ]
            )
        else:
            print("We expected at least one of x, y, z to be fixed, results are not guaranteed to work")
            vertices_arr = np.array(
                [
                    [min_x, min_y, max_z],
                    [min_x, max_y, min_z],
                    [max_x, max_y, min_z],
                    [max_x, min_y, max_z],
                ]
            )

        vertices = [Vertex.from_numpy(x) for x in vertices_arr]

        return Surface(vertices=vertices)

    def __init__(self, vertices, name=None):
        """Surface constructor."""
        if not isinstance(vertices, np.ndarray) and not isinstance(vertices, list):
            raise ValueError("Expected a list or numpy array of Vertex")

        for i, vertex in enumerate(vertices):
            if not isinstance(vertex, Vertex):
                raise ValueError(f"Element {i} is not a Vertex object")

        self.name = name
        self.os_plane = None

        self.vertices = copy.deepcopy(vertices)
        for vertex in self.vertices:
            vertex.surface = self

    def get_plane(self) -> Plane:
        """Returns the Plane of the Surface."""
        if self.os_plane is not None:
            return self.os_plane
        plane = openstudio.Plane(self.to_Point3dVector())
        self.os_plane = Plane(plane.a(), plane.b(), plane.c(), plane.d())
        return self.os_plane

    def get_plot_axis(self) -> str:
        """Returns a string representation of the plane it is on.

        TODO: raises if not exactly on 'xy', 'xz' or 'yz'
        """
        plane = self.get_plane()
        tol = 0.001
        if abs(abs(plane.a) - 1) < tol:
            return 'yz'
        if abs(abs(plane.b) - 1) < tol:
            return 'xz'
        if abs(abs(plane.c) - 1) < tol:
            return 'xy'

        # TODO
        raise NotImplementedError("Surface is not on a standard plane!")

    def plane(self) -> Plane:
        """Compute the plane from outwardNormal and the first point, not using OpenStudio."""
        normalVector = self.outwardNormal()
        if not np.isclose(normalVector.length(), 1.0):
            raise ValueError("Normal Unit Vector doesn't appear to be a unit vector")
        self.vertices[0]

        # d = -normalVector.x() * point.x() - normalVector.y() * point.y() - normalVector.z() * point.z();
        d = (-normalVector).dot(self.vertices[0])

        p = Plane(normalVector.x, normalVector.y, normalVector.z, d)
        for i, v in enumerate(self.vertices):
            if not p.pointOnPlane(v):
                print(f"Vertex {i} is not on the plane")
        return p

    def area(self) -> float:
        """Compute area of the surface."""
        newellVector = getNewellVector(self.vertices)
        return newellVector.length() / 2.0

    def outwardNormal(self) -> Vertex:
        """Returns the outward normal (normal unit vector)."""
        return getOutwardNormal(self.vertices)

    def tilt(self) -> float:
        """Returns the tilt of the surface, in radians, that is the angle between the outwardNormal and the Z axis."""
        z = Vertex(0.0, 0.0, 1.0)
        return getAngle(self.outwardNormal(), z)

    def azimuth(self) -> float:
        """Returns the azimuth of the surface, in radians.

        That is the angle between the outwardNormal and the North axis (Y-axis).
        """
        normal = self.outwardNormal()
        north = Vertex(0.0, 1.0, 0.0)
        angle = getAngle(normal, north)
        if normal.x < 0:
            return -angle + 2.0 * np.pi
        return angle

    def os_area(self) -> Vertex:
        """Returns area of the surface via openstudio."""
        return openstudio.getArea(self.to_Point3dVector()).get()

    def perimeter(self) -> float:
        """Returns the perimeter of the surface."""
        return sum([edge.length() for edge in self.to_Surface3dEdges()])

    def rough_centroid(self) -> Vertex:
        """Returns the centroid calculated in a rough way: the mean of the coordinates."""
        return Vertex.from_numpy(np.array([x.to_numpy() for x in self.vertices]).mean(axis=0))

    def os_centroid(self) -> Vertex:
        """Returns the centroid via openstudio."""
        centroid_ = openstudio.getCentroid(self.to_Point3dVector())
        if not centroid_.is_initialized():
            raise ValueError("OpenStudio failed to calculate centroid")
        return Vertex.from_Point3d(centroid_.get())

    def to_Point3dVector(self) -> List[openstudio.Point3d]:
        """Converts vertices to a list openstudio.Point3d."""
        return [v.to_Point3d() for v in self.vertices]

    def to_OSSurface(self, model: openstudio.model.Model) -> openstudio.model.Surface:
        """Creates an openstudio.model.Surface in the model passed as argument."""
        return openstudio.model.Surface(self.to_Point3dVector(), model)

    def to_numpy(self) -> np.ndarray:
        """Get a numpy array representing the vertices."""
        return np.array([v.to_numpy() for v in self.vertices])

    def to_Surface3dEdges(self) -> List[Surface3dEge]:
        """Converts vertex pairs to Surface3dEge."""
        edges = []
        for i, curVertex in enumerate(self.vertices):
            if i == len(self.vertices) - 1:
                nextVertex = self.vertices[0]
            else:
                nextVertex = self.vertices[i + 1]
            edges.append(Surface3dEge(start=curVertex, end=nextVertex, firstSurface=self))
        return edges

    def split_into_n_segments(self, n_segments, axis=None, plot=False) -> List[Surface]:
        """Splits a surface in N equal segments.

        If axis is not passed, it defaults to the first one of the plane
        eg: for a plane 'xy' it splits on 'x'
        """
        plot_axis = self.get_plot_axis()
        if axis is None:
            axis = plot_axis[0]
        if axis not in plot_axis:
            raise ValueError(f"This surface's plane is '{plot_axis}', so can't split on {axis=}")

        if n_segments < 2:
            raise ValueError("At least 2 segments needed")

        axis_to_index = {'x': 0, 'y': 1, 'z': 2}
        idx = axis_to_index[axis]
        v_np = self.to_numpy()
        minimum = v_np[:, idx].min()
        maximum = v_np[:, idx].max()
        segment_length = (maximum - minimum) / n_segments
        is_max = v_np[:, idx] == maximum
        is_min = ~is_max

        cur_min = minimum
        cur_max = cur_min + segment_length

        new_surfaces = []

        for i in range(n_segments):
            # print(cur_min, cur_max)
            v_np_i = v_np.copy()
            v_np_i[is_min, idx] = cur_min
            v_np_i[is_max, idx] = cur_max
            new_surface = Surface.from_numpy_array(v_np_i)
            if self.name:
                new_surface.name = f'{self.name}-{i+1}'
            new_surfaces.append(new_surface)

            cur_min = cur_max
            cur_max += segment_length

        if plot:
            fig, ax = plt.subplots(figsize=(16, 9))
            for new_surface in new_surfaces:
                new_surface.plot(ax=ax)

        return new_surfaces

    def rotate(self, degrees: float, axis=None) -> Surface:
        """Rotates a surface by an amount of degrees.

        Args:
        -----
        * degrees (float): the angle to rotate it by, in degrees. Positive means clockwise
        * axis (Vertex): if none, uses the Z axis

        Returns:
        ---------
        * a new Surface object with rotated vertices
        """
        if axis is None:
            axis = Vertex(0.0, 0.0, 1.0)

        # Lazy load to avoid circular import
        from geomeffibem.transformation import Transformation

        return Transformation.Rotation(axis=axis, radians=-openstudio.degToRad(degrees)) * self

    def translate(self, translation: Vertex) -> Surface:
        """Translates a surface along a translation vector."""
        from geomeffibem.transformation import Transformation

        return Transformation.Translation(translation=translation) * self

    def plot(self, name: Union[bool, str] = True, **kwargs):
        """Calls plot_vertices, cf help(plot_vertices)."""
        if isinstance(name, str):
            name = name
        elif name:
            name = self.name
        return plot_vertices(surface_like=self, name=name, **kwargs)

    def __repr__(self):
        """Repr."""
        s = ""
        if self.name is not None:
            s += f"Surface '{self.name}' = "
        s += "["
        if self.name is not None:
            s += "\n "
        imax = len(self.vertices) - 1
        for i, v in enumerate(self.vertices):
            if i > 0:
                s += " "
            s += f"{v}"
            if i < imax:
                s += ",\n"
        s += "]"
        return s


def get_surface_from_surface_like(surface_like: Union[Surface, List[Vertex], openstudio.model.Surface]) -> Surface:
    """Helper to get a Surface (class) from a surface like object."""
    if isinstance(surface_like, openstudio.model.Surface):
        surface = Surface.from_Surface(surface_like)
    elif isinstance(surface_like, Surface):
        surface = surface_like
    else:
        if isinstance(surface_like, list):
            surface_like = np.array(surface_like)

        if isinstance(surface_like[0], openstudio.Point3d):
            surface = Surface.from_Point3dVector(surface_like)
        elif isinstance(surface_like[0], np.ndarray):
            surface = Surface.from_numpy_array(surface_like)
        elif isinstance(surface_like[0], Vertex):
            surface = Surface(surface_like)

    return surface


def plot_vertices(
    surface_like: Union[Surface, List[Vertex], openstudio.model.Surface],
    ax=None,
    center_axes=False,
    with_rough_centroid=False,
    with_os_centroid=False,
    annotate=True,
    linewidth=None,
    force_align=False,
    name=None,
    plane=None,
    annotate_kwargs=dict(color='r', xytext=(5, 5), textcoords='offset points'),
    # Passed to ax.plot/plt.plot
    **kwargs,
):
    """Plot any surface-like object in 2D.

    Accepts a Surface, a list or numpy array of Vertex, or an openstudio.model.Surface object

    TODO: Assumes the surface is planar and falls exactly on 'xy', 'xz' or 'yz' plane currently
    """
    surface = get_surface_from_surface_like(surface_like=surface_like)

    is_aligned = False
    if plane is None and not force_align:
        try:
            plane = surface.get_plot_axis()
        except NotImplementedError as e:
            print(e)
            force_align = True
    if plane is None and force_align:
        print("aligning Face")
        # Lazy load to avoid circular import
        from geomeffibem.transformation import Transformation

        faceTransformation = Transformation.alignFace(surface.vertices)
        faceTransformationInverse = faceTransformation.inverse()
        points = faceTransformationInverse * surface.vertices
        surface = Surface(points, name=surface.name)
        plane = 'xy'
        is_aligned = True

    points = surface.to_numpy()

    if plane == 'xy':
        xs = points[:, 0]
        ys = points[:, 1]
    elif plane == 'xz':
        xs = points[:, 0]
        ys = points[:, 2]
    elif plane == 'yz':
        xs = points[:, 1]
        ys = points[:, 2]
    else:
        raise ValueError("plane must be in ['xy', 'xz', 'yz']")
    if ax is None:
        # print("Making a figure")
        max_width = xs.max() - xs.min()
        max_height = ys.max() - ys.min()
        h = 6
        w = h * max_width / max_height
        # print(w, h)
        fig, ax = plt.subplots(figsize=(w, h))

    ax.plot(np.append(xs, xs[0]), np.append(ys, ys[0]), marker='x', markeredgecolor='r', linewidth=linewidth, **kwargs)
    ax.set_xlabel(plane[0] if not is_aligned else "x'")
    ax.set_ylabel(plane[1] if not is_aligned else "y'")
    if annotate:
        for i, (x, y) in enumerate(zip(xs, ys)):
            ax.annotate(f"{i+1} ({x:.2f}, {y:.2f})", xy=(x, y), **annotate_kwargs)

    if with_rough_centroid:
        centroid_x, centroid_y = surface.rough_centroid().get_coords_on_plane(plane=plane)
        ax.annotate(f"rough ({centroid_x}, {centroid_y})", xy=(centroid_x, centroid_y))
        ax.plot(centroid_x, centroid_y, 'rx')
    if with_os_centroid or name is not None and name is not False:
        centroid_x, centroid_y = surface.os_centroid().get_coords_on_plane(plane=plane)
        if with_os_centroid:
            ax.annotate(f"os ({centroid_x}, {centroid_y})", xy=(centroid_x, centroid_y))
            ax.plot(centroid_x, centroid_y, 'gx')
            if name:
                ax.annotate(
                    name + (" (aligned)" if is_aligned else ""),
                    xy=(centroid_x, centroid_y),
                    xytext=(0, 50),
                    textcoords='offset pixels',
                    color='b',
                    arrowprops=dict(edgecolor='b', lw=1, ls='-', arrowstyle='->'),
                )
        else:
            ax.annotate(
                name + (" (aligned)" if is_aligned else ""), xy=(centroid_x, centroid_y), ha='center', va='center'
            )

    ax.spines['right'].set_color('none')
    ax.spines['top'].set_color('none')  # hide top axis

    if center_axes:
        ax.spines['bottom'].set_position('zero')  # x-axis where y=0
        ax.spines['left'].set_position('zero')
        ax.xaxis.set_label_position("top")

    return ax



"""A Polyhedron is a collection of Surface objects.

It's meant to represent a volume.
You can check whether it's enclosed or not.
"""
from __future__ import annotations

import copy
from typing import List, Tuple

import numpy as np

from geomeffibem.surface import Surface, Surface3dEge
from geomeffibem.vertex import Vertex, getNewellVector


class Polyhedron:
    """A collection of Surfaces, meant to represent a Volume."""

    def __init__(self, surfaces: List[Surface]):
        """Constructor from a list of Surface objects."""
        if not isinstance(surfaces, np.ndarray) and not isinstance(surfaces, list):
            raise ValueError("Expected a list or numpy array of Surfaces")

        for i, surface in enumerate(surfaces):
            if not isinstance(surface, Surface):
                raise ValueError(f"Element {i} is not a Surface object")
        self.surfaces = surfaces

    def get_surface_by_name(self, name):
        """Locate a surface by its name."""
        for s in self.surfaces:
            if s.name is not None and s.name == name:
                return s

    def numVertices(self):
        """Counts the total number of vertices for all surfaces."""
        count = 0
        for s in self.surfaces:
            count += len(s.vertices)
        return count

    def uniqueVertices(self) -> List[Vertex]:
        """Get a list of unique vertices (uses Vertex __eq__ operator which has a tolerance)."""
        uniqueVertices: List[Vertex] = []
        for s in self.surfaces:
            for vertex in s.vertices:
                found = False
                for unique_v in uniqueVertices:
                    if unique_v == vertex:
                        found = True
                        break
                if not found:
                    uniqueVertices.append(vertex)
        return uniqueVertices

    @staticmethod
    def edgesNotTwoForEnclosedVolumeTest(zonePoly: Polyhedron) -> Tuple[List[Surface3dEge], List[Surface3dEge]]:
        """Counts the number of times an Edge is used.

        Returns the ones that isn't used twice (and the ones used twice for debugging/inspection)
        """
        uniqueSurface3dEdges: List[Surface3dEge] = []

        for surface in zonePoly.surfaces:
            for edge in surface.to_Surface3dEdges():
                found = False
                for uniqEdge in uniqueSurface3dEdges:
                    if uniqEdge == edge:
                        uniqEdge.allSurfaces.append(surface)
                        found = True
                        break
                if not found:
                    uniqueSurface3dEdges.append(edge)

        edgesNotTwoCount = [x for x in uniqueSurface3dEdges if x.count() != 2]
        edgesTwoCount = [x for x in uniqueSurface3dEdges if x.count() == 2]
        return edgesNotTwoCount, edgesTwoCount

    def updateZonePolygonsForMissingColinearPoints(self) -> Polyhedron:
        """Creates a new Polyhedron with extra vertices when a point is found to be on a line segment."""
        updZonePoly = copy.deepcopy(self)

        uniqVertices = self.uniqueVertices()

        for surface in updZonePoly.surfaces:
            insertedVertex = True
            while insertedVertex:
                insertedVertex = False
                for i, edge in enumerate(surface.to_Surface3dEdges()):
                    for testVertex in uniqVertices:
                        if edge.containsPoints(testVertex):
                            if i == len(surface.vertices) - 1:
                                inext = 0
                            else:
                                inext = i + 1
                            surface.vertices.insert(inext, testVertex)
                            insertedVertex = True
                            break
                    # Break out of the loop on vertices/edges too, start again at while loop
                    if insertedVertex:
                        break
        return updZonePoly

    def isEnclosedVolume(self) -> Tuple[bool, List[Surface3dEge]]:
        """Checks if the Polyhedron is enclosed, that is all its edges are used exactly twice."""
        edgeNot2orig, _ = Polyhedron.edgesNotTwoForEnclosedVolumeTest(zonePoly=self)
        if not edgeNot2orig:
            return True, []

        print("Updating Polyhedron with collinear vertices on lines")
        updatedZonePoly = self.updateZonePolygonsForMissingColinearPoints()
        edgeNot2again, _ = Polyhedron.edgesNotTwoForEnclosedVolumeTest(updatedZonePoly)
        if not edgeNot2again:
            return True, []

        return False, edgesInBoth(edgeNot2orig, edgeNot2again)

    def calcPolyhedronVolume(self) -> float:
        """Calculates the Volume of an ENCLOSED Polyhedron."""
        volume = 0.0
        for surface in self.surfaces:
            vertices = surface.vertices
            p3FaceOrigin = vertices[1] - Vertex(0, 0, 0)
            newellAreaVector = getNewellVector(vertices)
            pyramidVolume = newellAreaVector.dot(p3FaceOrigin)
            volume += pyramidVolume
        # Our newellArea vector has twice the length
        volume /= 6.0
        return volume

    def to_os_cpp_code(self):
        """For my own convenience when writting OpenStudio tests."""
        for i, sf in enumerate(self.surfaces):
            if sf.name is not None:
                name = sf.name
            else:
                name = f"Surface {i+1}"
            if name[1] == '-':
                cleaned_name = name[2:].lower().replace('-', '')
            else:
                cleaned_name = name.lower().replace('-', '')
            s = "{"
            if sf.name is not None:
                s += "\n "
            n_vertices = len(sf.vertices)
            imax = n_vertices - 1
            for i, v in enumerate(sf.vertices):
                if i > 0:
                    s += " "
                s += f"{{{v.x:+.1f}, {v.y:+.1f}, {v.z:+.1f}}}"
                if i < imax:
                    s += ",\n"
            s += "}"

            print(f'Surface {cleaned_name}({s}, m);')
            print(f'{cleaned_name}.setName("{name}");')
            print(f'{cleaned_name}.setSpace(s);\n')

    def to_eplus_cpp_code(self):
        """For my own convenience when writting EnergyPlus tests."""
        n_surfaces = len(self.surfaces)
        print(
            f"""
            Array1D_bool enteredCeilingHeight;
            state->dataGlobal->NumOfZones = 1;
            enteredCeilingHeight.dimension(state->dataGlobal->NumOfZones, false);
            state->dataHeatBal->Zone.allocate(state->dataGlobal->NumOfZones);
            state->dataHeatBal->Zone(1).HasFloor = true;
            state->dataHeatBal->Zone(1).HTSurfaceFirst = 1;
            state->dataHeatBal->Zone(1).AllSurfaceFirst = 1;
            state->dataHeatBal->Zone(1).AllSurfaceLast = {n_surfaces};

            state->dataSurface->Surface.allocate({n_surfaces});
            """
        )
        for i, sf in enumerate(self.surfaces):
            n_vertices = len(sf.vertices)
            name = sf.name
            if 'ROOF' in name:
                tilt = 0.0
                c = 'SurfaceClass::Roof'
            elif 'FLOOR' in name:
                tilt = 180.0
                c = 'SurfaceClass::Floor'
            else:
                tilt = 90.0
                c = 'SurfaceClass::Wall'

            print(
                f'''
            state->dataSurface->Surface({i+1}).Name = "{name}";
            state->dataSurface->Surface({i+1}).Sides = {n_vertices};
            state->dataSurface->Surface({i+1}).Vertex.dimension({n_vertices});
            state->dataSurface->Surface({i+1}).Class = {c};
            state->dataSurface->Surface({i+1}).Tilt = {tilt};'''
            )

            for j, v in enumerate(sf.vertices):
                print(f"    state->dataSurface->Surface({i+1}).Vertex(1) = Vector({v.x}, {v.y}, {v.z});")


def edgesInBoth(a: List[Surface3dEge], b: List[Surface3dEge]) -> List[Surface3dEge]:
    """Helper function."""
    in_both = []
    for edge_a in a:
        for edge_b in b:
            if edge_a == edge_b:
                in_both.append(edge_a)
                break
    return in_both



"""Bounding Box."""

from typing import Optional, Tuple

import numpy as np

from geomeffibem.vertex import Vertex


class BoundingBox:
    """A crude BoundingBox.

    As you add points to it, the min/max x, y, z are updated.
    """

    def __init__(self):
        """Constructor for BoundingBox."""
        self.minX = None
        self.minY = None
        self.minZ = None
        self.maxX = None
        self.maxY = None
        self.maxZ = None

    def get_figsize(self, width=12) -> Tuple[float, float]:
        """Figure out a figure size (Broken)."""
        (x, y, z) = self.dimensions().to_numpy()
        return (width, width * x / y)

    def corners(self) -> Optional[np.ndarray]:
        """Returns an  of all 8 corner Points."""
        # TODO: make it return Vertex?
        if self.isEmpty():
            return None
        return np.array(
            [
                [self.minX, self.minY, self.minZ],
                [self.maxX, self.minY, self.minZ],
                [self.minX, self.maxY, self.minZ],
                [self.maxX, self.maxY, self.minZ],
                [self.minX, self.minY, self.maxZ],
                [self.maxX, self.minY, self.maxZ],
                [self.minX, self.maxY, self.maxZ],
                [self.maxX, self.maxY, self.maxZ],
            ]
        )

    def centerPoint(self) -> Vertex:
        """Get center of the bounding box, as the midpoint of the min/max coordinates."""
        if self.minX is None:
            raise ValueError("You need to add some points")

        return Vertex(x=(self.minX + self.maxX) / 2.0, y=(self.minY + self.maxY) / 2.0, z=(self.minZ + self.maxZ) / 2.0)

    def isEmpty(self) -> bool:
        """Checks if the bounding box is initialized."""
        return self.minX is None

    def addPoint(self, vertex) -> None:
        """Adds a single point and updates the min/max x, y, z."""
        if self.isEmpty():
            self.minX = vertex.x
            self.minY = vertex.y
            self.minZ = vertex.z
            self.maxX = vertex.x
            self.maxY = vertex.y
            self.maxZ = vertex.z
        else:
            self.minX = min(self.minX, vertex.x)
            self.minY = min(self.minY, vertex.y)
            self.minZ = min(self.minZ, vertex.z)

            self.maxX = max(self.maxX, vertex.x)
            self.maxY = max(self.maxY, vertex.y)
            self.maxZ = max(self.maxZ, vertex.z)

    def addPoints(self, vertices):
        """Adds multiple points and updates the min/max x, y, z."""
        [self.addPoint(v) for v in vertices]

    def dimensions(self) -> Vertex:
        """Returns the dimensions of the bounding box."""
        return Vertex(self.maxX - self.minX, self.maxY - self.minY, self.maxZ - self.minZ)



"""Vertex and Vertex related functions.

A Vertex is a 3-coordinate class, that can be used to represent a Point3d or a Vector3d
Includes utilities to go from/to numpy array and OpenStudio's Point3d
"""

from __future__ import annotations

from typing import Any, List, Tuple

import numpy as np
import openstudio


class Vertex:
    """Point3d and Vector3d."""

    @staticmethod
    def from_numpy(arr):
        """Factory method to construct from a numpy array (or a list) of 3 coords."""
        if not isinstance(arr, np.ndarray):
            arr = np.array(arr)
        if arr.shape != (3,):
            raise ValueError(f"Expected a numpy array with a dimension (3, ) (a Vector3d), got {arr.shape}")
        return Vertex(arr[0], arr[1], arr[2])

    @staticmethod
    def from_Point3d(pt: openstudio.Point3d):
        """Factory method to construct from an openstudio.Point3d."""
        return Vertex(pt.x(), pt.y(), pt.z())

    def __init__(self, x, y, z):
        """Vertex constructor."""
        self.x = x
        self.y = y
        self.z = z
        self.surface = None

    def copy(self):
        """Make a copy of this Vertex."""
        return Vertex(self.x, self.y, self.z)

    def get_coords_on_plane(self, plane='xy') -> Tuple[float, float]:
        """Returns two coordinates on a given plane."""
        if plane == 'xy':
            return self.x, self.y
        elif plane == 'xz':
            return self.x, self.z
        elif plane == 'yz':
            return self.y, self.z

        raise ValueError("Expected plane to be 'xy', 'xz' or 'yz'.")

    def length(self) -> float:
        """Get the length of the vector."""
        return np.sqrt(np.sum(self.to_numpy() ** 2))

    def normalize(self) -> Vertex:
        """Normalize to a length of 1, returns a copy."""
        v = self.copy()
        v.setLength(1.0)
        return v

    def setLength(self, newLength: float) -> None:
        """Change length of vector, in place."""
        currentLength = self.length()
        if currentLength > 0:
            mult = newLength / currentLength
            self.x *= mult
            self.y *= mult
            self.z *= mult
        else:
            raise ValueError("Cannot normalize a vector of length 0")

    def dot(self, other) -> float:
        """Computes the dot / scalar / inner product product of two vectors.

        (a.b).
        """
        # return np.dot(v.to_numpy(), v2.to_numpy())
        return self.x * other.x + self.y * other.y + self.z * other.z

    def cross(self, other, normalize: bool = False) -> Vertex:
        """Computes the cross product (a x b), which is a vector perpendicular to both a and b."""
        v = Vertex(
            x=(self.y * other.z - self.z * other.y),
            y=(self.z * other.x - self.x * other.z),
            z=(self.x * other.y - self.y * other.x),
        )
        if normalize:
            return v.normalize()
        return v

    def outer_product(self, other) -> np.ndarray:
        """Compute the outer product of this by another vector."""
        return np.outer(self.to_numpy(), other.to_numpy())

    def __add__(self, other) -> Vertex:
        """Return a + b."""
        return Vertex(x=self.x + other.x, y=self.y + other.y, z=self.z + other.z)

    def __neg__(self):
        """Return obj negated (-obj)."""
        return Vertex(-self.x, -self.y, -self.z)

    def __sub__(self, other) -> Vertex:
        """Return a - b."""
        return self + -other
        # return Vertex(
        #     x=self.x - other.x,
        #     y=self.y - other.y,
        #     z=self.z - other.z
        # )

    def __mul__(self, other) -> Vertex:
        """Multiplies each coordinate by a scalar."""
        if not isinstance(other, int) and not isinstance(other, float):
            raise ValueError("Multiplication of a vertex by something else than a numeric is not supported")
        return Vertex(self.x * other, self.y * other, self.z * other)

    def __rmul__(self, other) -> Vertex:
        """Multiplies each coordinate by a scalar."""
        if not isinstance(other, int) and not isinstance(other, float):
            raise ValueError("Multiplication of a vertex by something else than a numeric is not supported")
        return self.__mul__(other)

    def __truediv__(self, other) -> Vertex:
        """Divides each coordinate by a scalar."""
        if not isinstance(other, int) and not isinstance(other, float):
            raise ValueError("Division of a vertex by something else than a numeric is not supported")
        return Vertex(self.x / other, self.y / other, self.z / other)

    def __floordiv__(self, other) -> Vertex:
        """Divides each coordinate by a scalar."""
        if not isinstance(other, int) and not isinstance(other, float):
            raise ValueError("Division of a vertex by something else than a numeric is not supported")
        return Vertex(self.x // other, self.y // other, self.z // other)

    def to_numpy(self) -> np.ndarray:
        """Export to a numpy array of 3 coordinates."""
        return np.array([self.x, self.y, self.z])

    def to_Point3d(self) -> openstudio.Point3d:
        """Export to an openstudio.Point3d."""
        return openstudio.Point3d(self.x, self.y, self.z)

    def __eq__(self, other):
        """Operator equal. Raises if not passed a Vertex."""
        if not isinstance(other, Vertex):
            raise NotImplementedError("Not implemented for any other types than Vertex itself")
        return isAlmostEqual3dPt(self, other)

    def __ne__(self, other):
        """Operator not equal."""
        return not self == other

    def __repr__(self):
        """Repr."""
        # return f"({self.x}, {self.y}, {self.z})"
        return f"({self.x:+.4f}, {self.y:+.4f}, {self.z:+.4f})"

    # def __str__(self):
    #     return f"Vertex {self.__repr__()}"


def isAlmostEqual3dPt(v1: Vertex, v2: Vertex, tol=0.0127) -> bool:
    """Checks if both vertices almost equal within tolerance."""
    # 0.0127 m = 1.27 cm = 1/2 inch
    return not (abs((v1.to_numpy() - v2.to_numpy())) >= tol).any()


def distance(lhs: Vertex, rhs: Vertex) -> float:
    """Distance between two vertices."""
    squared_dist = np.sum((lhs.to_numpy() - rhs.to_numpy()) ** 2, axis=0)
    dist = np.sqrt(squared_dist)
    return dist


def distanceFromPointToLine(start: Vertex, end: Vertex, test: Vertex) -> np.floating[Any]:
    """Distance between a point and a line."""
    s = start.to_numpy()
    e = end.to_numpy()
    p = test.to_numpy()
    return np.linalg.norm(np.cross(e - s, p - s) / np.linalg.norm(e - s))


def isPointOnLineBetweenPoints(start: Vertex, end: Vertex, test: Vertex, tol: float = 0.0127) -> bool:
    """Checks whether a Vertex is on a segment.

    If the distance(start, test) + distance(test, end) == distance(start, end) then it's on a line
    But we first check that the distance from the point to the line is also inferior to this tolerance
    """
    if distanceFromPointToLine(start=start, end=end, test=test) < tol:
        return abs(distance(start, end) - (distance(start, test) + distance(test, end))) < tol
    return False


def getAngle(start: Vertex, end: Vertex) -> float:
    """Returns the angle between two vectors, in radians."""
    start = start.normalize()
    end = end.normalize()
    return np.arccos(start.dot(end))


def getNewellVector(points: List[Vertex]) -> Vertex:
    """Compute Newell vector from a list of points, direction is same as outward normal magnitude is twice the area."""
    n = len(points)
    if n < 3:
        raise ValueError("Cannot compute Newell Vector for less than 3 points")

    newellVector = Vertex(x=0, y=0, z=0)
    for i in range(n - 1):
        v1 = points[i] - points[0]
        v2 = points[i + 1] - points[0]
        newellVector += v1.cross(v2)

    return newellVector


def getOutwardNormal(points: list[Vertex]) -> Vertex:
    """Compute outward normal from a list of points."""
    newellVector = getNewellVector(points)
    return newellVector.normalize()



"""Plane."""

import numpy as np

from geomeffibem.vertex import Vertex, distance


class Plane:
    """A 3D Plane.

    Equation is `ax + by + cz + d = 0`
    """

    def __init__(self, a, b, c, d):
        """Plane constructor."""
        self.a = a
        self.b = b
        self.c = c
        self.d = d

    def outwardNormal(self) -> np.ndarray:
        """The outwardNormal of the plane."""
        return Vertex(self.a, self.b, self.c)

    def is_orthogonal(self) -> bool:
        """Checks if the plane is orthogonal."""
        return np.sum(np.abs(self.outwardNormal().to_numpy())) == 1.0

    def pointOnPlane(self, point: Vertex, tol=0.001) -> bool:
        """Checks whether the Vertex is on the Plane."""
        # project point to plane
        projected = self.project(point)

        return distance(point, projected) <= tol

    def project(self, point: Vertex) -> Vertex:
        """Project a point onto a Plane."""
        # http://www.9math.com/book/projection-point-plane
        if not isinstance(point, Vertex):
            raise ValueError("Expected a Vertex object")
        u = point.x
        v = point.y
        w = point.z

        num = self.a * u + self.b * v + self.c * w + self.d
        den = self.a * self.a + self.b * self.b + self.c * self.c  # this should always be 1.0
        ratio = num / den

        x = u - self.a * ratio
        y = v - self.b * ratio
        z = w - self.c * ratio

        return Vertex(x, y, z)

    def __repr__(self):
        """Repr."""
        return f"Plane ({self.a}, {self.b}, {self.c}, {self.d})"



#!/usr/bin/env python
"""Tests for `geomeffibem` Surface class."""

import numpy as np
import pytest

from geomeffibem.plane import Plane
from geomeffibem.surface import Surface
from geomeffibem.vertex import Vertex


def test_plane():
    """Tests we get the right plane from a Surface."""
    floor_surface = Surface.Floor(min_x=0.0, max_x=10.0, min_y=0.0, max_y=10.0, z=3.0)
    p = floor_surface.get_plane()
    assert p == floor_surface.get_plane()
    assert p.is_orthogonal()
    assert np.array_equal(p.outwardNormal().to_numpy(), np.array([-0.0, -0.0, -1.0]))
    assert p.a == 0.0
    assert p.b == 0.0
    assert p.c == -1.0
    assert p.d == 3.0


def test_project():
    """Test projection of a Vertex on a plane and pointOnPlane."""
    p = Plane(0.0, 0.0, -1.0, 3.0)
    v = Vertex(10.0, 10.0, 5.0)
    v2 = p.project(v)
    assert v2 == Vertex(10.0, 10.0, 3.0)
    assert not p.pointOnPlane(v)
    assert p.pointOnPlane(v2)
    with pytest.raises(ValueError):
        p.project(p)



"""Transformations: Rotation, Translations, and combinations of both."""

from __future__ import annotations

from typing import Optional

import numpy as np

from geomeffibem.plane import Plane
from geomeffibem.surface import Surface
from geomeffibem.vertex import Vertex, getOutwardNormal


class Transformation:
    """Transformation class."""

    @staticmethod
    def Rotation(axis: Vertex, radians: float, point: Optional[Vertex] = None) -> Transformation:
        """Constructs a Rotation Transformation (factory method)."""
        temp = axis.normalize()
        normalVector = temp.to_numpy()
        P = np.outer(normalVector, normalVector)

        # Rodrigues' rotation formula / Rotation matrix from Euler axis/angle
        # I*cos(radians) + I*(1-cos(radians))*axis*axis^T + Q*sin(radians)
        # Q = [0, -axis[2], axis[1]; axis[2], 0, -axis[0]; -axis[1], axis[0], 0]
        Q = np.zeros(shape=(3, 3))
        Q[0, 1] = -normalVector[2]
        Q[0, 2] = normalVector[1]
        Q[1, 0] = normalVector[2]
        Q[1, 2] = -normalVector[0]
        Q[2, 0] = -normalVector[1]
        Q[2, 1] = normalVector[0]

        c = np.cos(radians)
        identity_3 = np.identity(3)
        R = identity_3 * c + (1.0 - c) * P + Q * np.sin(radians)
        result = np.identity(4)
        result[:-1, :-1] = R
        if point is not None:
            #  translate point to origin, rotate, and then translate back
            # t_ori = Transformation.Translation(point)
            # t_rot = Transformation(result)
            # t_reverse = Transformation.Translation(-point)
            # return (t_ori, t_rot, t_reverse)
            return Transformation.Translation(point) * Transformation(result) * Transformation.Translation(-point)

        return Transformation(result)

    @staticmethod
    def Translation(translation: Vertex) -> Transformation:
        """Constructs a Translation Transformation (factory method)."""
        result = np.identity(4)
        result[0, 3] = translation.x
        result[1, 3] = translation.y
        result[2, 3] = translation.z
        return Transformation(result)

    # @staticmethod
    # def RotationAroundPoint(point: Vertex, axis: Vertex, radians) -> Transformation

    @staticmethod
    def alignZPrime(zPrime: Vertex) -> Transformation:
        """Transforms system with z' to regular system.

        will try to align y' with z, but if that fails will align y' with y
        """
        xp = None
        yp = None
        zp = zPrime.normalize()
        zAxis = Vertex(0.0, 0.0, 1.0)
        negXAxis = Vertex(-1.0, 0.0, 0.0)

        # check if face normal is up or down
        if abs(zp.dot(zAxis)) < 0.99:
            # not facing up or down, set yPrime along zAxis
            yp = zAxis - (zp * zp.dot(zAxis))
            yp = yp.normalize()
            xp = yp.cross(zp)
            print("Not facing up or down, set yPrime along zAxis")
        else:
            # facing up or down, set xPrime along -xAxis
            print("Facing up or down, set xPrime along -xAxis")
            xp = negXAxis - (zp * zp.dot(negXAxis))
            xp = xp.normalize()
            yp = zp.cross(xp)

        storage = np.identity(4)
        storage[0, 0] = xp.x
        storage[1, 0] = xp.y
        storage[2, 0] = xp.z
        storage[0, 1] = yp.x
        storage[1, 1] = yp.y
        storage[2, 1] = yp.z
        storage[0, 2] = zp.x
        storage[1, 2] = zp.y
        storage[2, 2] = zp.z
        return Transformation(matrix=storage)

    @staticmethod
    def alignFace(vertices: list[Vertex]) -> Transformation:
        """Transforms face coordinates to regular system, face normal will be z'.

        Will try to align y' with z, but if that fails will align y' with y
        face origin will be minimum point in x', y' and z'=0
        will return identity transformation if cannot compute plane for vertices
        """
        zPrime: Vertex = getOutwardNormal(vertices)
        align: Transformation = Transformation.alignZPrime(zPrime)
        alignedVertices = align.inverse() * vertices
        minX = alignedVertices[0].x
        minY = alignedVertices[0].y
        minZ = alignedVertices[0].z
        for vertex in alignedVertices:
            minX = min(minX, vertex.x)
            minY = min(minY, vertex.y)
            minZ = min(minZ, vertex.z)
        translate = Transformation.Translation(Vertex(minX, minY, minZ))
        return align * translate

    def __init__(self, matrix: np.ndarray = None):
        """Constructor for Transformation."""
        if matrix is None:
            matrix = np.identity(4)
        elif matrix.shape != (4, 4):
            raise ValueError(f"Expected a matrix of dimension (4, 4), got {matrix.shape}")
        self.matrix = matrix

    def rotationMatrix(self) -> np.ndarray:
        """Returns the rotation portion of the Transformation."""
        return self.matrix[:-1, :-1]

    def translation(self) -> Vertex:
        """Returns the translation portion of the Transformation."""
        return Vertex(self.matrix[0, 3], self.matrix[1, 3], self.matrix[2, 3])

    def inverse(self) -> Transformation:
        """Returns a transformation which is the inverse of this."""
        t = Transformation()
        t.matrix = np.linalg.inv(self.matrix)
        return t

    def __mul__(self, other):  # -> Union[Vertex, Transformation, np.ndarray, Surface, Plane]:
        """Multiplies self by other.

        Accepts various objects: Vertex, Transformation, List of Vertex, Surface, Plane.
        """
        if isinstance(other, Vertex):
            temp = np.matmul(self.matrix, np.append(other.to_numpy(), 1.0))
            return Vertex(temp[0], temp[1], temp[2])
        elif isinstance(other, Transformation):
            return Transformation(np.matmul(self.matrix, other.matrix))
        elif isinstance(other, np.ndarray) or isinstance(other, list):
            return np.array([self * v for v in other])
        elif isinstance(other, Surface):
            name = "Rotated unnamed"
            if other.name is not None:
                name = f"Rotated {name}"
            return Surface(vertices=[self * v for v in other.vertices], name=name)
        elif isinstance(other, Plane):
            # translate a point on the plane, just project (0,0,0)
            point = other.project(Vertex(0.0, 0.0, 0.0))

            # Get a point at outward normal
            outwardNormal = other.outwardNormal()
            refPoint = point + outwardNormal

            # Translate the two points and recompute the normal
            newPoint = self * point
            newRefPoint = self * refPoint
            newNormal = newRefPoint - newPoint

            d = (-newNormal).dot(newPoint)

            p = Plane(newNormal.x, newNormal.y, newNormal.z, d)

            return p
        else:
            raise ValueError(f"Not implemented for type {type(other)}")

    def __repr__(self):
        """Repr."""
        return f"Transformation:\n{self.matrix.__str__()}"



#!/usr/bin/env python
"""Tests for `geomeffibem` Polyhedron class."""

import openstudio
import pytest

from geomeffibem.polyhedron import Polyhedron, edgesInBoth
from geomeffibem.surface import Surface


@pytest.fixture
def zonePoly():
    """A fixture to create a Zone Polyhedron."""
    floor_surface = Surface.Floor(min_x=0.0, max_x=10.0, min_y=0.0, max_y=10.0, z=0.0)
    m = openstudio.model.Model()
    space = openstudio.model.Space.fromFloorPrint(floor_surface.to_Point3dVector(), 3.0, m).get()
    for sf in space.surfaces():
        if sf.surfaceType() == 'Floor':
            sf.setName("FLOOR")
        elif sf.surfaceType() == 'RoofCeiling':
            sf.setName("ROOF")
        else:
            group = sf.planarSurfaceGroup().get()
            site_transformation = group.siteTransformation()
            site_vertices = site_transformation * sf.vertices()
            site_outward_normal = openstudio.getOutwardNormal(site_vertices).get()
            north = openstudio.Vector3d(0.0, 1.0, 0.0)
            if site_outward_normal.x() < 0.0:
                azimuth = 360.0 - openstudio.radToDeg(openstudio.getAngle(site_outward_normal, north))
            else:
                azimuth = openstudio.radToDeg(openstudio.getAngle(site_outward_normal, north))

            if azimuth >= 315.0 or azimuth < 45.0:
                facade = "4-North"
            elif azimuth >= 45.0 and azimuth < 135.0:
                facade = "3-East"
            elif azimuth >= 135.0 and azimuth < 225.0:
                facade = "1-South"
            elif azimuth >= 225.0 and azimuth < 315.0:
                facade = "2-West"

            sf.setName(f"{facade}".upper())  # - Abs azimuth {azimuth:.2f}".upper())

    return Polyhedron([Surface.from_Surface(s) for s in m.getSurfaces()])


@pytest.fixture
def zonePolySplitWall(zonePoly):
    """A fixture to create a Zone Polyhedron with a split wall."""
    wall = zonePoly.get_surface_by_name('1-SOUTH')
    new_walls = wall.split_into_n_segments(n_segments=2, axis='x')

    return Polyhedron(surfaces=[x for x in zonePoly.surfaces if x.name != '1-SOUTH'] + new_walls)


def test_polyhedron(zonePoly):
    """Test a Polyhedron that is already enclosed in the first pass."""
    # This should work
    zonePoly.get_surface_by_name("1-SOUTH")
    # Six four-sided surfaces in a box
    assert zonePoly.numVertices() == 4 * 6

    isEnclosed, edgesNot2 = zonePoly.isEnclosedVolume()
    assert isEnclosed
    assert not edgesNot2


def test_polyhedron_split_wall(zonePolySplitWall):
    """Test a Polyhedron with a split wall, so it has to call updateZonePolygonsForMissingColinearPoints."""
    # This should work
    zonePolySplitWall.get_surface_by_name("1-SOUTH")
    # Six in a box, except one wall is split, so 7 four-sided surfaces
    assert zonePolySplitWall.numVertices() == 4 * 7

    isEnclosed, edgesNot2 = zonePolySplitWall.isEnclosedVolume()
    assert isEnclosed
    assert not edgesNot2


def test_polyhedron_not_enclosed(zonePoly):
    """Test a non enclosed volume."""
    # This should work
    zonePoly.surfaces.pop(0)
    # Six in a box, minus one, so 5 four-sided surfaces
    assert zonePoly.numVertices() == 4 * 5

    isEnclosed, edgesNot2 = zonePoly.isEnclosedVolume()
    assert isEnclosed is False
    assert edgesNot2


def test_cpp_generation(zonePoly):
    """Just trying to get coverage to ignore that stuff."""
    zonePoly.to_eplus_cpp_code()
    zonePoly.surfaces[0].name = None
    zonePoly.to_os_cpp_code()


def test_polyhedron_constructor():
    """Test the ctor error handling."""
    with pytest.raises(ValueError):
        Polyhedron(surfaces=5)

    with pytest.raises(ValueError):
        Polyhedron(surfaces=[5])


def test_edgesInBoth():
    """Test the helper edgesInBoth."""
    surface = Surface.Rectangle(min_x=0.0, max_x=10.0, min_y=0.0, max_y=10.0, min_z=0.0, max_z=0.0)
    surface.name = 'Floor'
    edges = surface.to_Surface3dEdges()
    assert len(edges) == 4
    assert len(edgesInBoth(edges, edges)) == 4
    assert len(edgesInBoth(edges, edges[:2])) == 2
    assert len(edgesInBoth(edges, [])) == 0
    assert len(edgesInBoth([], edges)) == 0



"""Top-level package for GeomEffiBEM."""

__author__ = """Julien Marrec"""
__email__ = 'contact@effibem.com'
__version__ = '0.1.6'

from geomeffibem.boundingbox import BoundingBox
from geomeffibem.plane import Plane
from geomeffibem.polyhedron import Polyhedron
from geomeffibem.surface import Surface, Surface3dEge, plot_vertices
from geomeffibem.transformation import Transformation
from geomeffibem.vertex import (
    Vertex,
    distance,
    distanceFromPointToLine,
    getAngle,
    getNewellVector,
    getOutwardNormal,
    isAlmostEqual3dPt,
    isPointOnLineBetweenPoints,
)
